﻿module PDFConvertLib

open iTextSharp.text
open iTextSharp.text.pdf
open System.Drawing
open System.Drawing.Imaging
open BitEffectCS
open System.IO

let outputPath = @"c:\temp\"
let outputTempPath = Path.Combine(outputPath, @"pdfout\")
let effectedTempPath = Path.Combine(outputPath, @"pdfeffected\")

type msgCallback = delegate of string -> unit
type endCallback = delegate of unit -> unit
type clipper = delegate of System.Drawing.Image -> System.Drawing.Rectangle


let tempFileFormat = @"Page_{0}_Image_{1}.jpeg"
let numberOfThreads = 8

//  指定ディレクトリ作成
let rec createDirectory (path:string) =
  if System.IO.Directory.Exists(path) then
    ()
  else
    let currdir = 
      path.LastIndexOf "\\" |>
      (fun x -> (0, x)) |>
      path.Substring
    createDirectory currdir
    System.IO.Directory.CreateDirectory(path) |> ignore

createDirectory outputTempPath
createDirectory effectedTempPath


// PDF作成関数群

let addOnePage (document:Document) aspectRatio pageNum = 
  let path = System.IO.Path.Combine(effectedTempPath, System.String.Format(tempFileFormat, pageNum, 1))
  let jpeg = iTextSharp.text.Image.GetInstance(path)
  jpeg.SetAbsolutePosition(0.0f, 0.0f)
  jpeg.ScaleToFit(PageSize.A4.Width, PageSize.A4.Width * aspectRatio)
  document.Add jpeg |> ignore
  document.NewPage()  |> ignore


let CreatePdf outputPath aspectRatio numOfPage = 
  let margin = 0.0f
  let document = new Document(new iTextSharp.text.Rectangle(0.0f, 0.0f, PageSize.A4.Width, PageSize.A4.Width * aspectRatio), margin, margin, margin, margin)
  PdfWriter.GetInstance(document, new System.IO.FileStream(outputPath, System.IO.FileMode.Create)) |> ignore
  document.Open ()
  List.iter (addOnePage document aspectRatio) [1..numOfPage]
  document.Close();

// JPEGのエンコーダ作成
let GetImageEncoder (imageType:string) =
  let imageType = imageType.ToUpperInvariant()
  let encs = System.Drawing.Imaging.ImageCodecInfo.GetImageEncoders()
  Array.fold (
    fun a (b:ImageCodecInfo) ->
      match a with
      | None -> if b.FormatDescription = imageType then Some(b) else None
      | a -> a
    ) None encs

let jpegEncoder = GetImageEncoder("JPEG")

let getPixelFormat v = 
  match v with
  | 1  -> PixelFormat.Format1bppIndexed
  
  | 8  -> PixelFormat.Format8bppIndexed
  | 16 -> PixelFormat.Format16bppArgb1555
  | 24 -> PixelFormat.Format24bppRgb
  | 48 -> PixelFormat.Format48bppRgb
  | _  -> PixelFormat.Format24bppRgb


let parms = new System.Drawing.Imaging.EncoderParameters(1)
parms.Param.[0] <- new System.Drawing.Imaging.EncoderParameter(System.Drawing.Imaging.Encoder.Compression, byte 12)

let saveImage outputPath pageNumber (img:byte[] option) =
  match img with
  | None -> ()
  | Some image ->
      use memStream = new System.IO.MemoryStream(image)
      memStream.Position <- 0L
      use img = System.Drawing.Image.FromStream(memStream)
      let path = System.IO.Path.Combine(outputPath, System.String.Format(tempFileFormat, pageNumber, 1))
      match jpegEncoder with
      | None -> ()
      | Some enc -> img.Save(path, enc, parms)

// 画像の縮小、クリッピング、ガンマ補正

let gamma = new FastBitmap()

let shrinkAndClipImage (pageSize:Size) (clip:clipper) (image:System.Drawing.Image) =
  let destImage = new Bitmap(pageSize.Width, pageSize.Height)
  let g = Graphics.FromImage(destImage)
  g.InterpolationMode <- System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic
  let drawRect = new Rectangle(0, 0, pageSize.Width, pageSize.Height)
  let clipRect = clip.Invoke image
  g.DrawImage(image, drawRect, clipRect, GraphicsUnit.Pixel)
  let img = gamma.BeginAccess destImage
  gamma.GammaChange(img, pageSize.Width, pageSize.Height)
  gamma.EndAccess(destImage, img)
  g.Dispose ()
  destImage


let effectImage (msgcb:msgCallback) outputPath (pageSize:Size) (clip:clipper) pageNumber = 
  let inPath = System.IO.Path.Combine(outputTempPath, System.String.Format(tempFileFormat, pageNumber, 1))
  use inJpeg = System.Drawing.Image.FromFile(inPath);
  let outPath = System.IO.Path.Combine(outputPath, System.String.Format(tempFileFormat, pageNumber, 1))
  let parms = new System.Drawing.Imaging.EncoderParameters(1) 
  parms.Param.[0] <- new System.Drawing.Imaging.EncoderParameter(System.Drawing.Imaging.Encoder.Compression, byte 0)
  match jpegEncoder with
  | None -> ()
  | Some enc ->
      use shrinked = shrinkAndClipImage pageSize clip inJpeg
      shrinked.Save(outPath, enc, parms)
      sprintf "%A %d" inPath pageNumber |>
      msgcb.Invoke

//  複数の資源を振り分ける
let ThreadRoundRobbin theMethod max cntThread n =
  async {
    List.iter theMethod [n..cntThread..max]
  }

//  画像にエフェクトをかけてPDFに保存


let EffectAndCreatePdfOperation outputPath (pageSize:Size) (msgcb:msgCallback) (endcb:endCallback)(clip:clipper) (gammaPara:float) (numPages:int) =
  async {
    gamma.InitGamma gammaPara
    List.map (ThreadRoundRobbin (effectImage msgcb effectedTempPath pageSize clip) numPages numberOfThreads) [1..numberOfThreads] |>
    Async.Parallel |>
    Async.RunSynchronously |> ignore
    CreatePdf outputPath ((float32 pageSize.Height) / (float32 pageSize.Width)) numPages
    endcb.Invoke ()
  }
 
let effectImages outputPath (pageSize:Size) (msgcb:msgCallback) (endcb:endCallback) (clip:clipper) (gammaPara:float) (numPages:int) =
  EffectAndCreatePdfOperation outputPath pageSize msgcb endcb clip gammaPara numPages |>
  Async.Start |> ignore
  
//  PDFオブジェクトからJPEGのオブジェクトを抽出
let getImage (doc:PdfReader) (theObj:PdfObject) =
  let tg = PdfReader.GetPdfObject(theObj) :?> PdfDictionary
  let subtype = PdfReader.GetPdfObject(tg.Get(PdfName.SUBTYPE)) :?> PdfName
  if PdfName.IMAGE.Equals(subtype) then
    let xrefIdx = (theObj :?> PRIndirectReference).Number
    let pdfObj = doc.GetPdfObject(xrefIdx)
    let str = pdfObj :?> PdfStream
    let filter = tg.Get(PdfName.FILTER).ToString()
    match filter with
    | "/FlateDecode" -> None
    | _ -> Some(PdfReader.GetStreamBytesRaw(str :?> PRStream))
  else
    None

let getPdfObjects (xobj:PdfDictionary) (doc:PdfReader) =
  seq {
    match xobj with
    | null -> ()
    | xobj ->
      for key in xobj.Keys do
        let theObj = xobj.Get(key)
        if theObj.IsIndirect() then
          yield getImage doc theObj
  }

let getImages outputPath (dict:PdfDictionary) (doc:PdfReader) page =
  let res = PdfReader.GetPdfObject(dict.Get(PdfName.RESOURCES)) :?> PdfDictionary
  let xobj = PdfReader.GetPdfObject(res.Get(PdfName.XOBJECT)) :?> PdfDictionary
  getPdfObjects xobj doc
  |> Seq.iter (saveImage outputPath page)


// PDFからイメージを抽出
let OnePageFrom (msgcb:msgCallback) outputPath (pdf:PdfReader) (page:int) =
  let pg = pdf.GetPageN(page)
  getImages outputPath pg pdf page
  sprintf "%A %d" pdf page |>
  msgcb.Invoke


let ExtractOperation (msgcb:msgCallback) (endcb:endCallback) (sourcePdf:string) outputPath =
  async {
    createDirectory outputPath
    let pdf = new PdfReader(sourcePdf)
//    List.iter (OnePageFrom msgcb outputPath pdf) [1..pdf.NumberOfPages]      //  この行を生かし、続く三行をコメントアウトすると、シングルスレッド処理になります
    List.map (ThreadRoundRobbin (OnePageFrom msgcb outputPath pdf) pdf.NumberOfPages numberOfThreads) [1..numberOfThreads] |>
    Async.Parallel |>
    Async.RunSynchronously |> ignore
    endcb.Invoke ()
  }

let ExtractImagesFromPDF (msgcb:msgCallback) (endcb:endCallback) (sourcePdf:string) outputPath =
  ExtractOperation msgcb endcb sourcePdf outputPath |>
  Async.Start |> ignore
